from __future__ import annotations
from typing import Type

import os
from pathlib import Path

from .base_config import BaseConfig, Field, SettingsConfigDict
from .system import System
from .workspace import Workspace, WorkspaceConfig
from .studio import Studio, StudioConfig
from .project import Project

_DEFAULT_SESSION: Session | None = None


class SessionConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_")
    verbose: bool = Field(False, description="Tchatty logs")


def set_default_session(session: Session):
    global _DEFAULT_SESSION
    _DEFAULT_SESSION = session
    # print("Default TGZR Session set.")
    env_home = os.environ.setdefault("tgzr_home", str(session.home))
    # print(f'  updated env "tgzr_home" to {env_home!r}')


def get_default_session() -> Session | None:
    """
    Return the session set as default with `set_default_session()`.
    If no session was set as default and the `tgzr_home` environment
    variable is set, a new session is created using it and set as
    default.
    Return None in other cases.
    """
    if _DEFAULT_SESSION is None:
        env_home = os.environ.get("tgzr_home") or os.environ.get("TGZR_HOME")
        if env_home is None:
            return None
        # print(f"Creating default TGZR session using 'tgzr_home' env var: {env_home}")
        session = Session(home=env_home)
        set_default_session(session)

    return _DEFAULT_SESSION


class Session:

    def __init__(self, home: Path | str | None = None):
        self._config_filename = ".tgzr"

        self._known_config_types: list[Type[BaseConfig]] = []
        self.declare_config_type(SessionConfig)
        self.declare_config_type(WorkspaceConfig)
        self.declare_config_type(StudioConfig)

        self._selected_studio_name: str | None = None
        self._selected_project_name: str | None = None

        self._system: System = System(self)
        self._workspace: Workspace | None = None

        if home is None:
            home = Path.cwd().resolve()
        else:
            home = Path(home)
        self.set_home(home)

    @property
    def home(self) -> Path:
        if self._home is None:
            raise ValueError("Home not set. Please use `set_home(path) first.`")
        return self._home

    def set_home(
        self, path: str | Path, ensure_config_found: bool = True
    ) -> Path | None:
        """Returns the path of the config loaded or None."""
        path = Path(path).resolve()
        config_path = SessionConfig.find_config_file(path, self._config_filename)
        if config_path is None:
            if ensure_config_found:
                raise FileNotFoundError(
                    f"Could not find a {self._config_filename} file under {path!r} and ancestors. "
                    "(Use `ensure_config_found=False` to bypass this error.)"
                )
            self._home = path
        else:
            self._home = config_path.parent
            self._config = SessionConfig.from_file(config_path)

        try:
            self._workspace = Workspace(self)
        except FileNotFoundError:
            self._workspace = None

        # If the given path is inside a studio, let's make this studio
        # the default one:
        if self._workspace and path.is_relative_to(self._workspace.path):
            studio_name = str(path)[len(str(self._workspace.path)) + 1 :].split(
                os.path.sep, 1
            )[0]
            try:
                self._workspace.set_default_studio(
                    studio_name,
                    ensure_exists=True,
                    save_config=False,  # Do not save, this is just for this session.
                )
            except FileNotFoundError:
                pass
        return config_path

    @property
    def system(self) -> System:
        return self._system

    @property
    def workspace(self) -> Workspace | None:
        return self._workspace

    @property
    def config(self) -> SessionConfig:
        if self._config is None:
            raise ValueError("Config not set. Please use `set_home(path) first.`")
        return self._config

    def save_config(self) -> Path:
        """Returns the path of the saved file."""
        return self.write_config_file(None, allow_overwrite=True)

    def write_config_file(
        self, path: str | Path | None, allow_overwrite: bool = False
    ) -> Path:
        """Returns the path of the saved file."""
        if path is None:
            path = self._home / self._config_filename
        path = Path(path).resolve()
        self._config.to_file(
            path,
            allow_overwrite=allow_overwrite,
            header_text="tgzr session config",
        )
        return path

    def declare_config_type(self, config_type: Type[BaseConfig]):
        """
        Declare the config type as used in the session.
        This is used for documentation and cli help.
        """
        self._known_config_types.append(config_type)

    def get_config_env_vars(self) -> list[tuple[str, list[tuple[str, str | None]]]]:
        """
        Returns a list of (name, description) for each config
        declared with `declare_config_type()`
        """
        config_env_vars = []
        for config_type in self._known_config_types:
            config_env_vars.append((config_type.__name__, config_type.used_env_vars()))
        return config_env_vars

    def set_verbose(self):
        self.config.verbose = True

    def set_quiet(self):
        self.config.verbose = False

    def select_studio(self, studio_name: str | None = None):
        self._selected_studio_name = studio_name

    @property
    def selected_studio_name(self) -> str | None:
        return self._selected_studio_name

    def get_selected_studio(self, ensure_exists: bool = True) -> Studio | None:
        if self.workspace is None:
            return None
        studio_name = self._selected_studio_name
        if studio_name is None:
            studio_name = self.workspace.default_studio_name()
        if studio_name is None:
            return None
        return self.workspace.get_studio(studio_name, ensure_exists)

    @property
    def selected_project_name(self) -> str | None:
        return self._selected_project_name

    def select_project(self, project_name: str | None = None):
        self._selected_project_name = project_name

    def get_selected_project(self) -> Project | None:
        if self.workspace is None:
            return None
        if self._selected_project_name is None:
            return None
        studio = self.get_selected_studio()
        if studio is None:
            return None
        return studio.get_project(self._selected_project_name)



import importlib.metadata
import os
import sys
import subprocess

import click

from tgzr.cli.utils import TGZRCliGroup

from tgzr.cli.add_plugins import BrokenCommand
from ..app_sdk.nice_app import ShellNiceApp
from ..app_sdk.qtpy_app import ShellQtpyApp
from ..app_sdk.exe_app import ShellExeApp


@click.group("app", cls=TGZRCliGroup, help="Run installed applications.")
def app_group():
    pass


def create_nice_run_command(app: ShellNiceApp) -> click.Command:
    @click.command(app.app_name)
    @click.option(
        "-r",
        "--reload",
        is_flag=True,
        help="Reload on code change (will run in browser).",
    )
    @click.option(
        "-d",
        "--detach",
        is_flag=True,
        help="Launch the App in another process to avoid blocking this one.",
    )
    def run_app(reload: bool, detach: bool):
        launcher_module = app.run_native_module.__name__
        if reload:
            if detach:
                raise click.UsageError(
                    "You cant run detached with reload (you wouldnt be able to close the process). Please choose one."
                )
            launcher_module = app.run_dev_module.__name__

        cmd = [sys.executable, "-m", launcher_module]
        click.echo(f"Running command: {' '.join(cmd)}")
        if detach:
            popen = subprocess.Popen(cmd)
            click.echo(f"New Process: {popen}")
        else:
            os.system(" ".join(cmd))

    return run_app


def create_run_command(app: ShellQtpyApp | ShellExeApp) -> click.Command:
    @click.command(app.app_name)
    @click.option(
        "-d",
        "--detach",
        is_flag=True,
        help="Launch the App in another process to avoid blocking this one.",
    )
    def run_app(detach: bool):
        launcher_module = app.run_native_module.__name__
        cmd = [sys.executable, "-m", launcher_module]
        click.echo(f"Running command: {' '.join(cmd)}")
        if detach:
            popen = subprocess.Popen(cmd)
            click.echo(f"New Process: {popen}")
        else:
            os.system(" ".join(cmd))

    return run_app


def install_plugin(group: click.Group):

    group.add_command(app_group)

    entry_point_group = "tgzr.shell.app_plugin"

    all_entry_points = importlib.metadata.entry_points(group=entry_point_group)

    for ep in all_entry_points:
        try:
            app = ep.load()
            if isinstance(app, ShellNiceApp):
                cmd = create_nice_run_command(app)
            else:
                cmd = create_run_command(app)
            app_group.add_command(cmd)
            app.cli_run_cmd_installed(cmd, group)
        except Exception as e:
            app_group.add_command(BrokenCommand(ep, e))
            raise

    return group



from typing import Literal

from pathlib import Path

import click
import os

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session
from ..workspace import Workspace
from ..studio import Studio

from .utils import pass_session


@click.group(cls=TGZRCliGroup)
def dev():
    """Developer tools"""
    pass


@dev.command()
@click.argument("index_name")
@pass_session
def create_local_index(session: Session, index_name: str):
    """
    Create a package index folder in *home*, and set it as default index on
    the workspace config.
    """
    workspace: Workspace = session.workspace

    pi = session.home / index_name
    pi.mkdir(exist_ok=True)
    workspace.config.default_index = "../" + index_name
    workspace.save_config()


@dev.command()
@click.option(
    "-i",
    "--index_name",
    help="Name of the local index. Default is to get it from Workspaces config.",
)
@click.option(
    "-p", "--project", help="Path of the project to build. Defaults to current path."
)
@click.option(
    "-B",
    "--builder",
    default="hatch",
    help='Builder to use, one of ["hatch", "uv", "build"] (defaults to "hatch").',
)
@pass_session
def build_to_local_index(
    session: Session,
    index_name: str | None,
    project: str | None,
    builder: Literal["hatch", "uv", "build"] = "hatch",
):
    """
    Build the project in the current directory directly in
    the specified local pacakge index (a folder in the *home*)

    Note: `hatch` must be installed, or you must use the --builder option.
    """
    workspace: Workspace = session.workspace

    if workspace is not None:
        index = workspace.resolve_index(index_name)
    else:
        index = workspaces.resolve_index(index_name)

    if index is None:
        raise click.UsageError(
            f"No index_name specified and no default index configured. Use --index-name or `tgzr ws dev create-local-index INDEX_NAME` first."
        )

    if "://" in index:
        raise click.UsageError(
            f'The index "{index}" is not a local path. Select another one with --index-name or create one with `tgzr ws dev create-local-index path/to/index`.'
        )

    local_index = Path(index)
    if not local_index.exists():
        raise click.UsageError(
            f'The folder "{index}" does not exists. Select another one with --index-name or create one with `tgzr ws dev create-local-index path/to/index` first.'
        )

    if builder == "hatch":
        cwd = None
        if project is not None:
            cwd = Path(".").resolve()
            os.chdir(project)
        cmd = f"hatch build {local_index}"
        try:
            os.system(cmd)
        finally:
            if cwd is not None:
                os.chdir(cwd)

    elif builder == "uv":
        src_option = ""
        if project is not None:
            src_option = project
        cmd = f"uv build --out-dir {local_index} {src_option}"
        os.system(cmd)

    elif builder == "build":
        src_option = ""
        if project is not None:
            src_option = project
        cmd = f"python -m build --outdir {local_index} {src_option}"
        os.system(cmd)

    else:
        raise click.UsageError(f"Unsupported builder: {builder!r}")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .session import Session


class System:
    """
    Utility class to manage the "System" folder in a tgzr session home.
    """

    def __init__(self, session: Session):
        self.session = session



from types import ModuleType

import os
import subprocess

from ._base_app import _BaseShellApp


class ShellExeApp(_BaseShellApp):
    def __init__(
        self,
        exe_path: str,
        app_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: list[str] = [],
    ):
        super().__init__(app_name, run_module, app_id, app_groups)
        self._exe_path = exe_path

    def exe_path(self) -> str:
        return self._exe_path

    def run_app(self):
        # TODO: find out how we can pass tgzr.cli args to here:
        os.system(self.exe_path())


class ShellHostApp(ShellExeApp):
    """
    This special type of application
    represents an "Integrated DCC".
    i.e: a DCC which can be controled
    by TGZR.
    """

    # TODO: implement this in a dedicated package.

    def __init__(
        self,
        host_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: list[str] = [],
    ):
        super().__init__(None, host_name, run_module, app_id, app_groups)

    def env(self) -> dict[str, str]:
        """
        Subclass will typically reimplement this to
        return the appropriate env dict.
        """
        return {}

    def exe_path(self) -> str:
        """
        Subclass will typically implement this to
        locate the appropriate exe path and return it.
        """
        raise NotImplementedError()

    def cmd(self) -> list[str]:
        """
        Subclass will typically reimplement this to
        build the full command, using self.exe_path().
        """
        return [self.exe_path()]

    def run_app(self):

        # TODO: implement a subprocess manager to keep track
        # of all popen.
        popen = subprocess.Popen(self.cmd(), env=self.env())



from __future__ import annotations
from types import ModuleType

from pathlib import Path
import inspect

import click

from tgzr.cli.utils import TGZRCliGroup

from .exceptions import MissingRunNativeModule


class _BaseShellApp:

    def __init__(
        self,
        app_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: list[str] = [],
    ):
        """
        Create an App which will be accessible in the tgzr CLI.

        `app_name`:
            The name of the app, used to name the CLI command.

        `app_id`:
            The unic id associated with the app.
            The default value is the name of the module defining the
            app. This should be unic enough.

        `run_module`:
            The python module used to run the app.
            Must be a valid modules with appropriate safeguards.
            NOTE: using None will raise an exception showing you
            how to implement this module.

        """
        frame: inspect.FrameInfo = inspect.stack()[1]
        if app_name is None:
            app_name = frame.function
            # print("---> AUTO APP NAME:", repr(app_name))

        if app_id is None:
            module = inspect.getmodule(frame[0])
            app_id = module and module.__name__ or "???"
            # print("---> AUTO APP ID:", app_id)

        if run_module is None:
            raise MissingRunNativeModule(self)
        self.run_native_module = run_module

        self.app_name = app_name
        self.app_id = app_id
        self.app_groups = app_groups

    def cli_run_cmd_installed(
        self, created_cmd: click.Command, root_group: TGZRCliGroup
    ):
        """
        Called when tgzr.shell.cli_plugin.app_cli has created and
        registered a cli command to execute this app.

        Subclasses can override this to alter the cmd or set it as default
        on the root group.

        Default does nothing.
        """
        pass

    def run_app(self):
        raise NotImplementedError()
